from django.conf.urls import url
from gfiles import views
# from django.contrib.auth import views as auth_views
from django.views.generic.edit import CreateView
from django.contrib.auth.forms import UserCreationForm


from django.urls import reverse_lazy
from django.contrib.auth.views import(
    LoginView,
    LogoutView
)




urlpatterns = [
    url(r'^$', views.index, name='index'),
    url(r'^login/$', LoginView.as_view(
        template_name='gfiles/login.html',
        success_url='/',
        redirect_authenticated_user=True
    ), name='login'),
    url(r'^logout/$', LogoutView.as_view(
        next_page=reverse_lazy('login'),
    ), name='logout'),
    url(r'^register/$', CreateView.as_view(
        template_name='gfiles/register.html',
        form_class=UserCreationForm,
        success_url='/'
    ), name='register'),
    url(r'^upload_gfile/$', views.GFileCreateView.as_view(), name='upload_gfile'),
    url(r'^gfile_summary/$', views.GFileListView.as_view(), name='gfile_summary'),
    url(r'^update/$', views.status_update, name='status_update'),
    url(r'^success/$', views.success, name='success'),
    url(r'^track_tasks/$', views.TrackTasksListView.as_view(), name='track_tasks'),
    url(r'^delete_track_task/(?P<pk>[\w]+)/$', views.TrackTasksDeleteView.as_view(), name='delete_track_task'),
    url(r'^track_task_progress/(?P<pk>\d+)/$', views.TrackTasksProgressView.as_view(), name='track_task_progress'),
]




# from sitetree.utils import tree, item
#
# # Be sure you defined `sitetrees` in your module.
# sitetrees = (
#   # Define a tree with `tree` function.
#   tree('gfiles_all', items=[
#       # Then define items and their children with `item` function.
#       item('GFILES', 'add_run', children=[
#           item('Upload generic file', 'upload_gfile', in_menu=True, in_sitetree=True),
#           item('Show all files', 'gfile_summary',  in_menu=True, in_sitetree=True),
#       ])
#   ]),
#   # ... You can define more than one tree for your app.
# )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.db import models
from django.contrib.auth.models import User
from datetime import datetime
import os


def data_file_store(instance, filename):
    now = datetime.now()
    return os.path.join('data', 'files', now.strftime("%Y_%m_%d_%H%M%S"), filename)


class GenericFile(models.Model):
    """ Model for managing generic files. Files can either be saved as symlink or copied to the data file store
    based on what has been set in the settings file MEDIA_ROOT parameter
    """
    data_file = models.FileField(upload_to=data_file_store, blank=True, null=True, max_length=500,
                                 help_text='The reference to the uploaded data file')
    original_filename = models.CharField(max_length=500, null=False, blank=False,
                                         help_text='If a file with the same filename is uploaded at the same time'
                                                   'a duplicate is saved with a unique name. This field is used'
                                                   'to keep track of the original name')
    user = models.ForeignKey(User, on_delete=models.CASCADE, help_text='The user who uploaded the file', null=True,
                             blank=True)

    def __str__(self):  # __unicode__ on Python 2
        return self.original_filename


class TrackTasks(models.Model):
    """ Model for tracking Celery tasks to users. If django-db is being used for the Celery backend then the
    taskid is already stored in the Django database. However this will not record the user, also this database table
    will be missing if redis or rabbitmq is used for the backend.
    """
    user = models.ForeignKey(User, on_delete=models.CASCADE, help_text='The user who uploaded the file', null=True,
                             blank=True)
    taskid = models.CharField(max_length=500, null=False, blank=False)

    state = models.CharField(max_length=500, null=True, blank=True)

    name = models.CharField(max_length=500, null=True, blank=True)

    result = models.URLField(null=True, blank=True)

    def __str__(self):  # __unicode__ on Python 2
        return '{}'.format(self.name)




# -*- coding: utf-8 -*-
#  python manage.py test misa -v 3 --keepdb
#  coverage run --source='.' manage.py test galaxy -v 3 --keepdb
#  coverage report
#  coverage html --omit="admin.py"
from __future__ import unicode_literals, print_function

import os
from django.core.files import File
from django.contrib.auth.models import AnonymousUser, User
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import TestCase, RequestFactory, Client
from django.urls import reverse

from gfiles.views import GFileCreateView, GFileListView
from gfiles.models import GenericFile

def add_middleware_to_request(request, middleware_class):
    middleware = middleware_class()
    middleware.process_request(request)
    return request

def add_middleware_to_response(request, middleware_class):
    middleware = middleware_class()
    middleware.process_response(request)
    return request


class UserAccountTestCase(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='jacob', email='jacob@…', password='top_secret')

    def test_signup(self):
        print(reverse('register'))

        response = self.client.post('/register/', {
            'username': 'admin99',
            'password1': 'qwertyui',
            'password2': 'qwertyui',
            'email': 'admin@example.com'})

        print(User.objects.all())
        print(response.status_code)
        print(response)
        user = User.objects.get(username="admin99")
        self.assertEqual(user.username, "admin99")
        self.assertEqual(response.status_code, 302)  # if it's successful it redirect


class GFileCreateViewTestCase(TestCase):

    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='jacob', email='jacob@…', password='top_secret')

        self.test_data_file_pth = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'tests', 'data',
                                          'test_data_file.tsv')

    def test_login_redirect(self):
        """
        Test to check if a guest user is redirect to the login page
        """
        request = self.factory.get(reverse('upload_gfile'))

        print(self.user)
        request.user = AnonymousUser()
        request = add_middleware_to_request(request, SessionMiddleware)
        response = GFileCreateView.as_view()(request)
        # print(response)

        # client acts as a fake website for the request
        response.client = Client()

        self.assertRedirects(response, '/login/?next=/upload_gfile/')

    def test_get(self):
        """
        """
        request = self.factory.get(reverse('upload_gfile'))
        request.user = self.user
        response = GFileCreateView.as_view()(request)
        self.assertEqual(response.status_code, 200)  # done without error

    def test_post_valid_url(self):
        """
        Test post with valid url
        """
        gfile_before = len(GenericFile.objects.all())

        with open(self.test_data_file_pth) as fp:
            self.client.force_login(self.user)
            response = self.client.post(reverse('upload_gfile'),
                                    {'data_file': fp})
            self.assertEqual(response.status_code, 302)  # redirected page (e.g. success page)
            print(GenericFile.objects.all())
            self.assertEqual(len(GenericFile.objects.all()), gfile_before + 1)



class GFileListViewTestCase(TestCase):

    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='jacob', email='jacob@…', password='top_secret')
        self.test_data_file_pth = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
                                               'tests', 'data', 'test_data_file.tsv')
        self.gfile = GenericFile(original_filename='test_data_file.tsv', user=self.user)
        self.gfile.data_file.save('test_data_file.tsv', File(open(self.test_data_file_pth, 'r')))

    def test_get_anom(self):
        """
        Any user should be able to see the table
        """
        request = self.factory.get(reverse('gfile_summary'))

        print(self.user)
        request.user = AnonymousUser()
        request = add_middleware_to_request(request, SessionMiddleware)
        response = GFileListView.as_view()(request)
        # print(response)

        # client acts as a fake website for the request
        response.client = Client()

        self.assertEqual(response.status_code, 200)  # done without error

    def test_get_table_entry(self):
        """
        Check elements in table
        """
        request = self.factory.get(reverse('gfile_summary'))

        print(self.user)
        request.user = AnonymousUser()
        request = add_middleware_to_request(request, SessionMiddleware)
        response = GFileListView.as_view()(request)
        # print(response)

        # client acts as a fake website for the request
        response.client = Client()

        self.gfile = GenericFile(original_filename='test_data_file.tsv', user=self.user)
        self.gfile.data_file.save('test_data_file.tsv', File(open(self.test_data_file_pth, 'r')))


        self.assertContains(response, '<div class="table-container">')
        self.assertContains(response, '<div id="GFileTableWithCheck" class="column-shifter-container">')
        self.assertContains(response, '<table  class="paleblue">')
        self.assertContains(response, '<td class="filename">test_data_file.tsv</td>')




import django_tables2 as tables
from .models import GenericFile, TrackTasks
from django_tables2_column_shifter.tables import ColumnShiftTable
from django_tables2.utils import A

TABLE_CLASS = "mogi table-bordered table-striped table-condensed table-hover"

class GFileTable(ColumnShiftTable):
    """ Class for django-tables2 table of :class:`gfiles.models.GenericFiles`.

    Inherits the ColumnShiftTable, that is a modified django-tables2 class that allows columns to be dynamically
    selected
    """
    filename = tables.Column(accessor='original_filename', verbose_name='Filename')

    class Meta:
        model = GenericFile
        attrs = {"class": TABLE_CLASS}
        fields = ('id', )


class GFileTableWithCheck(GFileTable):
    """ Class for django-tables2 table of :class:`gfiles.models.GenericFiles` with a check box column

    Inherits :class:`gfiles.models.GFileTable` and just adds checkbox.
    """
    check = tables.CheckBoxColumn(accessor="pk",
                                           attrs={
                                               "th__input": {"onclick": "toggle(this)"},
                                               "td__input": {"onclick": "addfile(this)"}},
                                           orderable=False)

    class Meta:
        model = GenericFile
        attrs = {"class": "table table-bordered table-striped table-condensed table-hover", }
        fields = ('id', 'filename', 'user', 'check')


class TrackTasksTable(ColumnShiftTable):
    progress = tables.LinkColumn('track_task_progress', text='progress', verbose_name='Monitor Progress', args=[A('pk')])
    delete = tables.LinkColumn('delete_track_task', text='delete', verbose_name='Delete', args=[A('pk')])
    class Meta:
        model = TrackTasks
        attrs = {"class": TABLE_CLASS}
        fields = ('id', 'user', 'name', 'taskid', 'state', 'result')









import os
import tempfile
from django.core.files import File
from django.conf import settings

def save_as_symlink(abs_pth, name, generic_file_obj):
    """ Takes a :class:`gfiles.models.GenericFiles` object and saves a symlink of the file to the absolute path
     (abs_pth).

    :param abs_pth: path to the file to symlink to
    :param name: Name to save to the file as
    :param generic_file_obj: :class:`gfiles.models.GenericFiles` object

    :returns: updated :class:`gfiles.models.GenericFiles` object with symlink

    """
    tf = tempfile.NamedTemporaryFile(delete=False)

    # first create link to empty file
    generic_file_obj.data_file.save(name, File(open(tf.name)))

    fpth = os.path.join(settings.MEDIA_ROOT, generic_file_obj.data_file.name)
    os.remove(fpth)
    os.symlink(abs_pth, fpth)
    generic_file_obj.save()
    tf.close()

    return generic_file_obj


# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-08-22 09:47
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import gfiles.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='GenericFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('data_file', models.FileField(blank=True, help_text='The reference to the uploaded data file', max_length=500, null=True, upload_to=gfiles.models.data_file_store)),
                ('original_filename', models.CharField(help_text='If a file with the same filename is uploaded at the same timea duplicate is saved with a unique name. This field is usedto keep track of the original name', max_length=500)),
                ('user', models.ForeignKey(blank=True, help_text='The user who uploaded the file', null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='TrackTasks',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('taskid', models.CharField(max_length=500)),
                ('state', models.CharField(blank=True, max_length=500, null=True)),
                ('name', models.CharField(blank=True, max_length=500, null=True)),
                ('result', models.URLField(blank=True, null=True)),
                ('user', models.ForeignKey(blank=True, help_text='The user who uploaded the file', null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.views.generic import CreateView, DetailView, DeleteView, UpdateView, ListView, View
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib import messages
import os
import copy
from django.contrib import messages
from django.utils.safestring import mark_safe

from django.http import JsonResponse
from celery.result import AsyncResult

from django_filters.views import FilterView
from django_tables2.views import SingleTableMixin, MultiTableMixin

from gfiles.models import GenericFile, TrackTasks
from gfiles.forms import GFileForm
from django.shortcuts import render
from django.views.generic import View

from gfiles.filter import GFileFilter, TrackTasksFilter
from gfiles.tables import GFileTableWithCheck, TrackTasksTable

from django_tables2.export.views import ExportMixin
from django.urls import reverse_lazy
from celery.task.control import revoke

class GFileCreateView(LoginRequiredMixin, CreateView):
    """ Class to create a save a generic file using the GenericFile model.

    Inherits the CreateView class and uses the LoginRequiredMixin
    """
    model = GenericFile
    success_msg = "File uploaded"
    success_url = reverse_lazy('success')
    form_class = GFileForm
    template_name = 'gfiles/gfile_form.html'

    def update_form(self, form):
        ofn = self.request.FILES['data_file'].name
        form.instance.original_filename = ofn
        form.instance.user = self.request.user
        return form

    def form_valid(self, form):
        messages.info(self.request, self.success_msg)
        form = self.update_form(form)
        return super(GFileCreateView, self).form_valid(form)


class GFileListView(ExportMixin, SingleTableMixin, FilterView):
    """ Class to view a table and filter all of the currently saved GenericFiles

    Inherits the FilterView class and uses the SingleTableMixin for viewing the django-tables2 table and
    uses the ExportMixin so that the table can be exported as a csv file
    """
    table_class = GFileTableWithCheck
    model = GenericFile
    template_name = 'gfiles/gfile_summary.html'
    filterset_class = GFileFilter


class TrackTasksListView(LoginRequiredMixin, SingleTableMixin, FilterView):
    """ Class to view a table and filter all of the currently saved GenericFiles

    Inherits the FilterView class and uses the SingleTableMixin for viewing the django-tables2 table and
    uses the ExportMixin so that the table can be exported as a csv file
    """
    table_class = TrackTasksTable
    model = TrackTasks
    template_name = 'gfiles/tracktask_summary.html'
    filterset_class = TrackTasksFilter

    def get_queryset(self):
        if not self.request.user.is_authenticated():
            return self.model.objects.none()
        qs = self.model.objects.filter(user=self.request.user)
        return qs


class TrackTasksProgressView(LoginRequiredMixin, View):
    """
    """
    def get(self, request, *args, **kwargs):


        tt = TrackTasks.objects.filter(pk=self.kwargs['pk'])

        if tt:
            request.session['result'] = tt[0].taskid

        return render(request, 'gfiles/status.html', {'s': 0, 'progress': 0})


class TrackTasksDeleteView(DeleteView):
    model = TrackTasks
    success_url = reverse_lazy('track_tasks')
    template_name = 'gfiles/confirm_delete.html'

    def post(self, request, *args, **kwargs):
        tt = self.get_object()
        revoke(tt.taskid, terminate=True)
        return self.delete(request, *args, **kwargs)

def async_task_progress(id):
    # https://blog.miguelgrinberg.com/post/using-celery-with-flask

    # Task has finished and information has been removed (i think)
    try:
        task = AsyncResult(id)
    except KeyError as e:
        print(e)
        # check if still have info in our database table regarding this task
        tt = TrackTasks.objects.filter(taskid=id)
        if tt:
            response = {
                'state': 'COMPLETED',
                'current': 1,
                'total': 1,
                'status': 'Task {} has completed'.format(tt[0].name),
            }
        else:
            response = {
                'state': 'REMOVED',
                'current': 1,
                'total': 1,
                'status': 'Task has been removed',
            }

        return JsonResponse(response)

    info = copy.deepcopy(task.info)  # incase things change or rabbit/redis deletes message
    if task.state == 'SUCCESS':
        response = {
            'state': task.state,
            'current': 1,
            'total': 1,
            'status': info['status'] if info and 'status' in info else '',
        }
        tt = TrackTasks.objects.filter(taskid=id)
        if tt and tt[0].result:
            response['status'] = mark_safe('<p><a href="{}">View result</a></p>'.format(tt[0].result))


    elif task.state == 'PENDING':
        print('pending')
        # job did not start yet
        response = {
            'state': task.state,
            'current': 0,
            'total': 1,
            'status': 'Pending...'
        }
    elif task.state != 'FAILURE':

        response = {
            'state': task.state,
            'current': info['current'],
            'total': info['total'],
            'status': info['status'],
        }
        # if 'result' in info:
        #     response['result'] = info
    else:
        # something went wrong in the background job
        response = {
            'state': task.state,
            'current': 0,
            'total': 0,
            'status': 'FAILURE (unknown)',
        }



    if response['state']=='FAILURE-KNOWN':
        response['state'] = 'FAILURE'

    if response['state'] == 'FAILURE':
        tt = TrackTasks.objects.filter(taskid=id)
        if tt:
            tt[0].state = 'FAILURE'
            tt[0].save()

    if task.state == 'SUCCESS':
        response['progress'] = 100
    elif task.info:
        if float(response['total']) == 0:
            response['progress'] = 0
        else:
            response['progress'] = (float(response['current']) / float(response['total'])) * 100.0
    else:
        response['progress'] = 0

    return response

def status_update(request):
    """ Updates for tracking status of long processes via celery
    """

    id = request.session['result']
    response = async_task_progress(id)

    return JsonResponse(response)

# class DeleteMessages(LoginRequiredMixin, View):
#     """
#     """
#     def get(self, request, *args, **kwargs):
#
#         messages = get_messages(request)
#         for msg in messages:
#             del msg
#
#         next = request.POST.get('next', '/')
#         return HttpResponseRedirect(next)





def index(request):
    """ basic index view
    """
    return render(request, 'gfiles/index.html')


def success(request):
    """ basic success view
    """
    return render(request, 'gfiles/success.html')

def handler404(request):
    return render(request, 'gfiles/404.html', status=404)

def handler500(request):
    return render(request, 'gfiles/500.html', status=500)


import django_filters
from .models import GenericFile


class GFileFilter(django_filters.FilterSet):
    """ Class for filtering Generic Files using django_filters.FilterSet

    Filters available for the username of the user who uploaded the file and the original filename (the original name
    can be different to the name of the file that is stored if multiple files of the same name are uploaded at the same
    time)
    """
    username = django_filters.CharFilter('user__username', lookup_expr='contains', label="Username")
    original_filename = django_filters.CharFilter('original_filename', lookup_expr='contains', label="Filename")

    class Meta:
        model = GenericFile
        fields = ('original_filename', 'username')


class TrackTasksFilter(django_filters.FilterSet):
    """ Class for filtering Tracked tasks
    """
    username = django_filters.CharFilter('user__username', lookup_expr='contains', label="Username")
    taskid = django_filters.CharFilter('taskid', lookup_expr='contains', label="taskid")
    state = django_filters.CharFilter('state', lookup_expr='contains', label="status")

    class Meta:
        model = GenericFile
        fields = ('username', 'taskid', 'state')



from django.utils.safestring import mark_safe

EYE = mark_safe('<i class="fa fa-eye" aria-hidden="true"></i>')
DOWN = mark_safe('<i class="fa fa-fw fa-download"></i>')
PLAY = mark_safe('<i class="fa fa-fw fa-play"></i>')
SAVE = mark_safe('<i class="fa fa-fw fa-save"></i>')


from django import forms
from gfiles.models import GenericFile


class GFileForm(forms.ModelForm):
    """ Form class for :class:`gfiles.models.GenericFiles
    """
    class Meta:
        model = GenericFile
        fields = ['data_file']